# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division, print_function, absolute_import

from typing import Optional, List, Tuple, Callable, Any
from functools import reduce
import numpy as np
import functools
import torch
import torch.nn.functional as F
import nnef
import math


# Helpers


_numpy_dtype_to_torch = {
    np.int8: torch.int8,
    np.int16: torch.int16,
    np.int32: torch.int32,
    np.int64: torch.int64,
    np.uint8: torch.uint8,
    np.float_: torch.float,
    np.double: torch.double,
    np.float16: torch.float16,
    np.float32: torch.float32,
    np.float64: torch.float64,
    np.int_: torch.int,
    np.short: torch.short,
    np.longlong: torch.long,
    np.bool_: torch.bool,
    int: torch.int,
    bool: torch.bool,
}


def _clamp(x, a, b):
    return max(a, min(b, x))


def _expand_to_rank(input, rank):
    # type: (torch.Tensor, int)->torch.Tensor
    rank_diff = rank - len(input.shape)
    return input.reshape(tuple(input.shape) + rank_diff * (1,))


def _expand_binary(input1, input2):
    # type: (torch.Tensor, torch.Tensor)->Tuple[torch.Tensor, torch.Tensor]
    rank = max(len(input1.shape), len(input2.shape))
    return _expand_to_rank(input1, rank), _expand_to_rank(input2, rank)


def _binary(f):
    def g(x, y):
        x, y = _expand_binary(x, y)
        return f(x, y)

    return g


def _prod(items):
    return functools.reduce(lambda x, y: x * y, items, 1)


def _same_padding(input, filter, stride, dilation):
    assert len(input) == len(filter) == len(stride) == len(dilation)

    output = [(ui + (s - 1)) // s for ui, s in zip(input, stride)]
    dilated = [(f - 1) * d + 1 for f, d in zip(filter, dilation)]
    total = [max(0, (di - 1) * s + df - ui) for di, s, df, ui in zip(output, stride, dilated, input)]

    return [(pad // 2, (pad + 1) // 2) for pad in total]


def _inverse_permutation(perm):
    inverse = [0] * len(perm)
    for i, p in enumerate(perm):
        inverse[p] = i
    return inverse


def _apply_permutation(items, perm):
    return [items[ind] for ind in perm]


# Operations

def _positive_pad(input, padding, border='constant', value=0.0):
    # type: (torch.Tensor, List[Tuple[int, int]], str, float)->torch.Tensor

    assert all(p >= 0 and q >= 0 for p, q in padding), "Negative padding is not supported "

    assert padding
    assert len(input.shape) in (3, 4, 5)
    assert padding[:2] == [(0, 0), (0, 0)] or (padding[0] == (0, 0) and padding[-1] == (0, 0))
    assert border in ("constant", "reflect", "replicate")

    rank = len(input.shape)
    needs_transpose = padding[0] == (0, 0) and padding[1] != (0, 0) and padding[-1] == (0, 0)
    if needs_transpose:
        padding = padding[1:-1]
        input = input.permute([0, rank - 1] + list(range(1, rank - 1)))
    else:
        padding = padding[2:]

    pad = []
    for p, q in reversed(padding):
        pad += [p, q]

    padded = F.pad(input=input, pad=pad, mode=border, value=value) if not all(p == 0 for p in pad) else input

    if needs_transpose:
        padded = padded.permute([0] + list(range(2, rank)) + [1])

    return padded


def nnef_pad(input, padding, border='constant', value=0.0):
    # type: (torch.Tensor, List[Tuple[int, int]], str, float)->torch.Tensor

    assert padding, \
        "nnef.pad does not support empty list as padding"
    assert len(input.shape) in (3, 4, 5), \
        "nnef.pad is only implemented for 3D, 4D, 5D tensors; got: {}D.".format(len(input.shape))
    assert padding[:2] == [(0, 0), (0, 0)] or (padding[0] == (0, 0) and padding[-1] == (0, 0)), \
        "nnef.pad is not implemented in N, C dimensions; got: {}.".format(padding)

    if all(p <= 1 and q <= 1 for p, q in padding) and border == "reflect-even":
        border = "replicate"

    assert border in ("constant", "reflect", "replicate"), \
        "nnef.pad is only implemented with 'constant', 'reflect' and 'replicate' border; got: {}.".format(border)

    input = _positive_pad(input,
                          padding=[(p if p > 0 else 0, q if q > 0 else 0) for p, q in padding],
                          border=border,
                          value=value)

    return nnef_slice(input,
                      axes=list(range(len(input.shape))),
                      begin=[-p if p < 0 else 0 for p, _q in padding],
                      end=[q if q < 0 else 0 for _p, q in padding])


nnef_add = _binary(lambda x, y: x + y)


def nnef_add_n(values):
    return nnef_add(values[0], nnef_add_n(values[1:])) if len(values) > 1 else values[0]


def nnef_conv(input,  # type: torch.Tensor
              filter,  # type: torch.Tensor
              bias,  # type: torch.Tensor
              border='constant',  # type: str
              padding=None,  # type: Optional[List[Tuple[int, int]]]
              stride=None,  # type: Optional[List[int]]
              dilation=None,  # type: Optional[List[int]]
              groups=1,  # type: int
              ):
    # type: (...)->torch.Tensor

    assert len(input.shape) in (3, 4, 5), "nnef.conv is only implemented for 3D, 4D, 5D tensors, given: {}D.".format(len(input.shape))

    bias = bias.reshape(1, 1).expand((1, filter.shape[0])) if _prod(bias.size()) == 1 else bias

    spatial_dims = len(input.shape[2:])
    groups = input.shape[1] if groups == 0 else groups
    stride = [1] * spatial_dims if not stride else stride
    dilation = [1] * spatial_dims if not dilation else dilation
    if not padding:
        padding = _same_padding(input=input.shape[2:],
                                filter=filter.shape[2:],
                                stride=stride,
                                dilation=dilation)

    pad = nnef_pad(input=input, padding=[(0, 0)] * 2 + padding, border=border)
    conv = {1: F.conv1d, 2: F.conv2d, 3: F.conv3d}[spatial_dims](input=pad,
                                                                 weight=filter,
                                                                 bias=bias.squeeze(dim=0).contiguous(),
                                                                 stride=tuple(stride),
                                                                 padding=0,
                                                                 dilation=tuple(dilation),
                                                                 groups=groups)

    return conv


def nnef_deconv(input,  # type: torch.Tensor
                filter,  # type: torch.Tensor
                bias,  # type: torch.Tensor
                border='constant',  # type: str
                padding=None,  # type: Optional[List[Tuple[int, int]]]
                stride=None,  # type: Optional[List[int]]
                dilation=None,  # type: Optional[List[int]]
                output_shape=None,  # type: Optional[List[int]]
                groups=1,  # type: int
                ):
    # type: (...)->torch.Tensor

    assert border == 'constant' or border == 'replicate', "nnef.deconv: '{}' border unsupported.".format(border)

    if output_shape and output_shape[0] != input.shape[0]:
        output_shape = list(output_shape)
        output_shape[0] = input.shape[0]

    rank = len(input.shape)
    assert rank in (3, 4, 5), "nnef.deconv is only implemented for 3D, 4D, 5D tensors, given: {}D.".format(len(input.shape))

    spatial_dims = len(input.shape[2:])
    stride = [1] * spatial_dims if not stride else stride
    dilation = [1] * spatial_dims if not dilation else dilation

    if groups == 0:
        if output_shape:
            groups = output_shape[1]
        else:
            # Planewise deconvolution without output_size, assuming that #(input channels) = #(output channels)
            groups = filter.shape[0]

    output_channels = filter.shape[1] * groups
    if output_shape:
        assert output_shape[1] == output_channels
    else:
        output_shape = nnef.shapes.deconv_shape(input=list(input.shape),
                                                filter=filter.shape,
                                                bias=bias.shape,
                                                border=border,
                                                padding=padding,
                                                stride=stride,
                                                dilation=dilation,
                                                groups=groups,
                                                output_shape=None)
    if not padding:
        padding = _same_padding(input=output_shape[2:],
                                filter=filter.shape[2:],
                                stride=stride,
                                dilation=dilation)

    if border == 'replicate':
        input = F.pad(input=input, pad=(1,) * 2 * spatial_dims, mode='replicate')
        padding = [(p + s, q + s) for (p, q), s in zip(padding, stride)]

    uncropped_output_shape = nnef.shapes.deconv_shape(input=list(input.shape),
                                                      filter=filter.shape,
                                                      bias=bias.shape,
                                                      border=border,
                                                      padding=[(0, 0)] * (rank - 2),
                                                      stride=stride,
                                                      dilation=dilation,
                                                      groups=groups,
                                                      output_shape=None)

    crop_before = [p for p, _q in padding]
    crop_after = [uncropped - out - before
                  for uncropped, out, before
                  in zip(uncropped_output_shape[2:], output_shape[2:], crop_before)]

    bias = bias.reshape(1, 1).expand((1, output_channels)) if _prod(bias.size()) == 1 else bias

    deconv = {1: F.conv_transpose1d,
              2: F.conv_transpose2d,
              3: F.conv_transpose3d}[spatial_dims](input=input,
                                                   weight=filter,
                                                   bias=bias.squeeze(dim=0).contiguous(),
                                                   stride=tuple(stride),
                                                   padding=0,
                                                   output_padding=0,
                                                   groups=groups,
                                                   dilation=tuple(dilation))

    return nnef_pad(deconv, padding=[(0, 0), (0, 0)] + [(-cb, -ca) for cb, ca in zip(crop_before, crop_after)])


def _evaluate_max_pool_or_box_params(input_shape, size, padding, stride, dilation):
    rank = len(input_shape)
    stride = [1] * rank if not stride else stride
    dilation = [1] * rank if not dilation else dilation
    padding = _same_padding(input=input_shape,
                            filter=size,
                            stride=stride,
                            dilation=dilation) if not padding else padding
    return padding, stride, dilation


def _max_pool_impl(input,  # type: torch.Tensor
                   size,  # type: List[int]
                   border='constant',  # type: str
                   padding=None,  # type: Optional[List[Tuple[int, int]]]
                   stride=None,  # type: Optional[List[int]]
                   dilation=None,  # type: Optional[List[int]]
                   with_index=False,  # type: bool
                   ):
    # type: (...)->torch.Tensor

    spatial_dims = len(input.shape) - 2
    value = float('-inf') if border == 'ignore' else 0.0
    border = 'constant' if border == 'ignore' else border

    pad = nnef_pad(input=input, padding=padding, border=border, value=value)

    result = {1: F.max_pool1d, 2: F.max_pool2d, 3: F.max_pool3d}[spatial_dims](input=pad,
                                                                               kernel_size=size[2:],
                                                                               stride=stride[2:],
                                                                               padding=0,
                                                                               dilation=dilation[2:],
                                                                               return_indices=with_index)
    return result


def _box_impl(input,  # type: torch.Tensor
              size,  # type: List[int]
              border,  # type: str
              padding,  # type: List[Tuple[int, int]]
              stride,  # type: List[int]
              dilation,  # type: List[int]
              normalize,  # type: bool
              ):
    # type: (...)->torch.Tensor

    assert 3 <= len(input.shape) <= 5
    assert len(input.shape) == len(size) == len(padding) == len(stride) == len(dilation)
    assert padding[:2] == [(0, 0), (0, 0)]
    assert size[:2] == stride[:2] == dilation[:2]

    assert not dilation or all(d == 1 for d in dilation), \
        "nnef.box (avg or sum pooling) is only implemented for dilation = 1."

    spatial_dims = len(input.shape) - 2

    pad = nnef_pad(input=input, padding=padding, border='constant' if border == 'ignore' else border)

    avg_pool = {1: F.avg_pool1d, 2: F.avg_pool2d, 3: F.avg_pool3d}[spatial_dims](
        input=pad,
        kernel_size=size[2:],
        stride=stride[2:],
        padding=0)

    if border == 'ignore' and normalize:
        ones = torch.ones_like(input)
        padded_ones = nnef_pad(input=ones, padding=padding, border='constant')
        avg_pool_ones = {1: F.avg_pool1d, 2: F.avg_pool2d, 3: F.avg_pool3d}[spatial_dims](
            input=padded_ones,
            kernel_size=size[2:],
            stride=stride[2:],
            padding=0)
        # If padding is big, zero averages can happen on the border, don't divide by zero
        avg_pool_ones = nnef_select(avg_pool_ones > 0, avg_pool_ones, torch.ones_like(avg_pool_ones))
        avg_pool /= avg_pool_ones

    if normalize:
        return avg_pool
    else:
        return avg_pool * _prod(size)


def _get_transform_for_box_or_max_pool(input_shape, active):
    # type: (List[int], List[bool])->Any
    assert len(input_shape) >= 3
    assert len(input_shape) == len(active)
    assert sum(active) <= 3, \
        "Sliding window operations are not supported if they have more than 3 'active' dimensions; got {}".format(sum(active))

    if 3 <= len(input_shape) <= 5 and not active[0] and not active[1]:  # Direct support
        return None, None, None, None
    else:
        inactive_dims = [i for i, a in enumerate(active) if not a]
        active_dims = [i for i, a in enumerate(active) if a]
        inactive_shape = [s for i, s in enumerate(input_shape) if i not in active_dims]
        active_shape = [s for i, s in enumerate(input_shape) if i in active_dims]
        perm = inactive_dims + active_dims
        perm_inv = _inverse_permutation(perm)
    return perm, perm_inv, inactive_shape, active_shape


def _box_or_max_pool(input,  # type: torch.Tensor
                     size,  # type: List[int]
                     border='constant',  # type: str
                     padding=None,  # type: Optional[List[Tuple[int, int]]],
                     stride=None,  # type: Optional[List[int]],
                     dilation=None,  # type: Optional[List[int]]
                     normalize=False,  # type: bool
                     is_max_pool=False,  # type: bool
                     ):
    assert not (normalize and is_max_pool)

    rank = len(input.shape)
    padding, stride, dilation = _evaluate_max_pool_or_box_params(input_shape=list(input.shape),
                                                                 size=size,
                                                                 padding=padding,
                                                                 stride=stride,
                                                                 dilation=dilation)
    active = [size_ != 1 or padding_ != (0, 0) or stride_ != 1 or dilation_ != 1
              for size_, padding_, stride_, dilation_
              in zip(size, padding, stride, dilation)]

    if sum(active) == 0:
        return input

    if rank < 3:
        perm, perm_inv, inactive_shape, active_shape = None, None, None, None
    else:
        perm, perm_inv, inactive_shape, active_shape = _get_transform_for_box_or_max_pool(list(input.shape), active)

    if rank < 3:
        input = input.unsqueeze(0).unsqueeze(0)
        size = [1, 1] + size
        padding = [(0, 0), (0, 0)] + padding
        stride = [1, 1] + stride
        dilation = [1, 1] + dilation
    elif perm is not None:
        input = input.permute(*perm)
        size = _apply_permutation(size, perm)
        padding = _apply_permutation(padding, perm)
        stride = _apply_permutation(stride, perm)
        dilation = _apply_permutation(dilation, perm)

        active_rank = len(active_shape)
        input = input.reshape(*[_prod(inactive_shape), 1] + active_shape)
        size = [1, 1] + size[-active_rank:]
        padding = [(0, 0), (0, 0)] + padding[-active_rank:]
        stride = [1, 1] + stride[-active_rank:]
        dilation = [1, 1] + dilation[-active_rank:]

    if is_max_pool:
        output = _max_pool_impl(
            input=input, size=size, border=border, padding=padding, stride=stride, dilation=dilation, with_index=False)
    else:
        output = _box_impl(input=input,
                           size=size,
                           border=border,
                           padding=padding,
                           stride=stride,
                           dilation=dilation,
                           normalize=normalize)

    if rank < 3:
        output = output.squeeze(0).squeeze(0)
    elif perm is not None:
        active_rank = len(active_shape)
        output = output.reshape(inactive_shape + list(output.shape)[-active_rank:])
        output = output.permute(*perm_inv)

    return output


def nnef_max_pool(input,  # type: torch.Tensor
                  size,  # type: List[int]
                  border='constant',  # type: str
                  padding=None,  # type: Optional[List[Tuple[int, int]]]
                  stride=None,  # type: Optional[List[int]]
                  dilation=None,  # type: Optional[List[int]]
                  ):
    # type: (...)->torch.Tensor
    return _box_or_max_pool(
        input, size=size, border=border, padding=padding, stride=stride, dilation=dilation, is_max_pool=True)


def nnef_max_pool_with_index(input,  # type: torch.Tensor
                             size,  # type: List[int]
                             border='constant',  # type: str
                             padding=None,  # type: Optional[List[Tuple[int, int]]]
                             stride=None,  # type: Optional[List[int]]
                             dilation=None,  # type: Optional[List[int]]
                             ):
    # type: (...)->torch.Tensor

    input_shape = list(input.shape)
    padding, stride, dilation = _evaluate_max_pool_or_box_params(input_shape=input_shape,
                                                                 size=size,
                                                                 padding=padding,
                                                                 stride=stride,
                                                                 dilation=dilation)

    assert len(input_shape) in (3, 4, 5), \
        "nnef.max_pool_with_index is only implemented for 3D, 4D, 5D tensors, given: {}D".format(len(input_shape))
    assert size[:2] == [1, 1], \
        "nnef.max_pool_with_index is only implemented for size = 1 in N and C dimensions"
    assert padding[:2] == [(0, 0), (0, 0)],\
        "nnef.max_pool_with_index is only implemented for padding = (0, 0) in N and C dimensions."
    assert stride[:2] == [1, 1], \
        "nnef.max_pool_with_index is only implemented for stride = 1 in N and C dimensions"
    assert dilation[:2] == [1, 1], \
        "nnef.max_pool_with_index is only implemented for dilation = 1 in N and C dimensions"

    return _max_pool_impl(input, size=size, border=border, padding=padding, stride=stride, dilation=dilation,
                          with_index=True)


def nnef_argmax_pool(input,  # type: torch.Tensor
                     size,  # type: List[int]
                     border='constant',  # type: str
                     padding=None,  # type: Optional[List[Tuple[int, int]]]
                     stride=None,  # type: Optional[List[int]]
                     dilation=None,  # type: Optional[List[int]]
                     ):
    # type: (...)->torch.Tensor
    _, index = nnef_max_pool_with_index(
        input, size=size, border=border, padding=padding, stride=stride, dilation=dilation)
    return index


def nnef_box(input,  # type: torch.Tensor
             size,  # type: List[int]
             border='constant',  # type: str
             padding=None,  # type: Optional[List[Tuple[int, int]]]
             stride=None,  # type: Optional[List[int]]
             dilation=None,  # type: Optional[List[int]]
             normalize=False,  # type: bool
             ):
    # type: (...)->torch.Tensor
    return _box_or_max_pool(
        input, size=size, border=border, padding=padding, stride=stride, dilation=dilation, normalize=normalize)


def nnef_debox(input,  # type: torch.Tensor
               size,  # type: List[int]
               border='constant',  # type: str
               padding=None,  # type: Optional[List[Tuple[int, int]]]
               stride=None,  # type: Optional[List[int]]
               dilation=None,  # type: Optional[List[int]]
               output_shape=None,  # type: Optional[List[int]]
               normalize=False,  # type: bool
               ):
    assert border in ('constant', 'ignore'), \
        "nnef.debox: '{}' border unsupported".format(border)
    assert len(size) in (3, 4, 5), \
        "nnef.debox is only implemented for 3D, 4D, 5D tensors, given: {}D".format(len(size))
    assert size[:2] == [1, 1], \
        "nnef.debox is only implemented for size = 1 in N and C dimensions"
    assert not padding or padding[:2] == [(0, 0), (0, 0)], \
        "nnef.debox is only implemented for padding = (0, 0) in N and C dimensions"
    assert not stride or stride[:2] == [1, 1], \
        "nnef.debox is only implemented for stride = 1 in N and C dimensions"
    assert not dilation or dilation[:2] == [1, 1], \
        "nnef.debox is only implemented for dilation = 1 in N and C dimensions"

    filter = torch.full(size=[input.shape[1], 1] + list(size)[2:],
                        fill_value=(1.0 / _prod(size) if normalize else 1.0),
                        device=input.device,
                        dtype=input.dtype)
    bias = torch.zeros(size=tuple(), device=input.device, dtype=input.dtype)

    return nnef_deconv(input=input,
                       filter=filter,
                       bias=bias,
                       border='constant',
                       padding=padding[2:] if padding else padding,
                       stride=stride[2:] if stride else stride,
                       dilation=dilation[2:] if dilation else dilation,
                       output_shape=output_shape,
                       groups=input.shape[1])


def nnef_avg_pool(input,  # type: torch.Tensor
                  size,  # type: List[int]
                  border='constant',  # type: str
                  padding=None,  # type: Optional[List[Tuple[int, int]]],
                  stride=None,  # type: Optional[List[int]],
                  dilation=None,  # type: Optional[List[int]]
                  ):
    # type: (...)->torch.Tensor
    return nnef_box(input, size=size, border=border, padding=padding, stride=stride, dilation=dilation, normalize=True)


def nnef_rms_pool(input,  # type: torch.Tensor
                  size,  # type: List[int]
                  border='constant',  # type: str
                  padding=None,  # type: Optional[List[Tuple[int, int]]],
                  stride=None,  # type: Optional[List[int]],
                  dilation=None,  # type: Optional[List[int]]
                  ):
    # type: (...)->torch.Tensor
    return torch.sqrt(nnef_avg_pool(torch.pow(input, 2.0),
                                    size=size,
                                    border=border,
                                    padding=padding,
                                    stride=stride,
                                    dilation=dilation))


def nnef_desample(input,  # type: torch.Tensor
                  index,  # type: torch.Tensor
                  size,  # type: List[int]
                  border='constant',  # type: str
                  padding=None,  # type: Optional[List[Tuple[int, int]]]
                  stride=None,  # type: Optional[List[int]]
                  dilation=None,  # type: Optional[List[int]]
                  output_shape=None,  # type: Optional[List[int]]
                  ):
    # type: (...)->torch.Tensor

    if output_shape and output_shape[0] != input.shape[0]:
        output_shape = list(output_shape)
        output_shape[0] = input.shape[0]

    input_shape = list(input.shape)
    index_shape = list(index.shape)
    rank = len(input_shape)
    spatial_dims = len(input_shape[2:])

    assert len(input_shape) in (3, 4, 5), \
        "nnef.desample is only implemented for 3D, 4D, 5D tensors, given: {}D".format(len(input_shape))
    assert not size or size[:2] == [1, 1], \
        "nnef.desample is only implemented for size = 1 in N and C dimensions"
    assert not padding or padding[:2] == [(0, 0), (0, 0)], \
        "nnef.desample is only implemented for padding = (0, 0) in N and C dimensions"
    assert not stride or stride[:2] == [1, 1], \
        "nnef.desample is only implemented for stride = 1 in N and C dimensions"
    assert not dilation or all(d == 1 for d in dilation), \
        "nnef.desample is only implemented for dilation = 1"

    stride = [1] * rank if not stride else stride
    dilation = [1] * rank if not dilation else dilation

    if not padding:
        calculated_output_shape = [i * s for i, s in zip(input_shape, stride)]
        padding = _same_padding(input=calculated_output_shape,
                                filter=size,
                                stride=stride,
                                dilation=dilation)
    else:
        calculated_output_shape = nnef.shapes.desample_shape(input_shape, index_shape,
                                                             size=size, border=border, padding=padding,
                                                             stride=stride, dilation=dilation, output_shape=None)

    output_shape = output_shape if output_shape else calculated_output_shape
    padded_output_shape = [s + p + q for s, (p, q) in zip(output_shape, padding)]
    unpooled = {1: F.max_unpool1d, 2: F.max_unpool2d, 3: F.max_unpool3d}[spatial_dims](
        input=input, indices=index, kernel_size=size[2:], stride=stride[2:], padding=0, output_size=padded_output_shape)
    return nnef_slice(unpooled,
                      axes=list(range(rank)),
                      begin=[p for p, _q in padding],
                      end=[p + s for (p, _q), s in zip(padding, output_shape)])


def nnef_batch_normalization(input,  # type: torch.Tensor
                             mean,  # type: torch.Tensor
                             variance,  # type: torch.Tensor
                             offset,  # type: torch.Tensor
                             scale,  # type: torch.Tensor
                             epsilon,  # type: float
                             is_training=False,  # type: bool
                             momentum=0.1,  # type: float
                             ):
    # type: (...)->torch.Tensor

    if isinstance(mean, torch.nn.Parameter):
        mean.requires_grad = False
    if isinstance(variance, torch.nn.Parameter):
        variance.requires_grad = False

    return F.batch_norm(input=input,
                        running_mean=nnef_squeeze(mean, axes=[0]),
                        running_var=nnef_squeeze(variance, axes=[0]),
                        weight=nnef_squeeze(scale, axes=[0]),
                        bias=nnef_squeeze(offset, axes=[0]),
                        training=is_training,
                        momentum=momentum,
                        eps=epsilon)


def _upsample_weights_1d(factor, symmetric):
    if symmetric:
        weights = [1 - (i + 0.5) / factor for i in range(factor)]
        weights = list(reversed(weights)) + weights
    else:
        weights = [1 - abs(i) / float(factor) for i in range(-factor + 1, factor)]
    return np.array(weights)


def _upsample_weights_2d(factor, symmetric):
    w0 = _upsample_weights_1d(factor[0], symmetric)
    w1 = _upsample_weights_1d(factor[1], symmetric)
    return np.outer(w0, w1)


def _upsample_weights_nd(factor, symmetric):
    ws = [_upsample_weights_1d(f, symmetric) for f in factor]
    return reduce(np.multiply, np.ix_(*ws))


def nnef_multilinear_upsample(input, factor, method='symmetric', border='replicate'):
    # type: (torch.Tensor, List[int], str, str)->torch.Tensor

    rank = len(factor)
    assert len(input.shape) == rank + 2

    mode = 'linear' if rank == 1 else 'bilinear'

    if method == 'aligned':
        return F.interpolate(input=input, scale_factor=tuple(factor), mode=mode, align_corners=True)
    elif method == 'symmetric' and border == 'replicate':
        return F.interpolate(input=input, scale_factor=tuple(factor), mode=mode, align_corners=False)

    n, c, = input.shape[:2]

    symmetric = method == 'symmetric'
    replicate = border == 'replicate'
    weights = _upsample_weights_nd(factor, symmetric)
    weights = np.tile(np.reshape(weights, newshape=(1, 1) + weights.shape), reps=(c, 1) + (1,) * rank)
    filter = torch.from_numpy(weights).to(device=input.device, dtype=input.dtype)
    bias = torch.zeros(size=tuple(), device=input.device, dtype=input.dtype)

    output_shape = [n, c] + [f * s for f, s in zip(factor, input.shape[2:])]

    if symmetric:
        return nnef_deconv(input, filter, bias, stride=factor, padding=[(f - 1, f - 1) for f in factor],
                           border='constant', groups=c, output_shape=output_shape)
    else:
        if replicate:
            input = nnef_pad(input, padding=[(0, 0), (0, 0)] + [(1, 0)] * rank, border=border)

        padding = factor if replicate else [f // 2 for f in factor]
        return nnef_deconv(input, filter, bias, stride=factor, padding=[(p, p) for p in padding],
                           border='constant', groups=c, output_shape=output_shape)


def nnef_nearest_upsample(input, factor):
    # type: (torch.Tensor, List[int])->torch.Tensor

    assert len(input.shape) in (3, 4, 5), \
        "nnef.nearest_upsample is only implemented for 3D, 4D, 5D tensors, given: {}D.".format(len(input.shape))

    return F.interpolate(input=input, scale_factor=tuple(factor), mode='nearest')


def nnef_softmax(x, axes=None):
    # type: (torch.Tensor, Optional[List[int]])->torch.Tensor

    axes = [1] if axes is None else axes

    if len(axes) == 0:
        return x
    elif len(axes) == 1:
        return F.softmax(x, dim=axes[0])
    else:
        m = nnef_max_reduce(x, axes=axes)
        e = torch.exp(x - m)
        return e / nnef_sum_reduce(x, axes=axes)


def nnef_local_response_normalization(input, size, alpha=1.0, beta=0.5, bias=1.0):
    # type: (torch.Tensor, List[int], float, float, float)->torch.Tensor

    sigma = bias + alpha * nnef_box(torch.pow(input, 2.0), size=size, normalize=True)
    return input / torch.pow(sigma, beta)


def nnef_local_mean_normalization(input, size):
    # type: (torch.Tensor, List[int])->torch.Tensor
    mean = nnef_box(input, size=size, normalize=True)
    return input - mean


def nnef_local_variance_normalization(input, size, bias=0.0, epsilon=0.0):
    # type: (torch.Tensor, List[int], float, float)->torch.Tensor
    sigma = torch.sqrt(nnef_box(torch.pow(input, 2.0), size=size, normalize=True))
    return input / torch.max(sigma + bias,
                             torch.full(size=[], fill_value=epsilon, device=input.device, dtype=input.dtype))


def nnef_local_contrast_normalization(input, size, bias=0.0, epsilon=0.0):
    # type: (torch.Tensor, List[int], float, float)->torch.Tensor
    centered = nnef_local_mean_normalization(input, size=size)
    return nnef_local_variance_normalization(centered, size=size, bias=bias, epsilon=epsilon)


def nnef_l1_normalization(input, axes, bias=0.0, epsilon=0.0):
    # type: (torch.Tensor, List[int], float, float)->torch.Tensor
    sigma = nnef_sum_reduce(torch.abs(input), axes=axes)
    return input / torch.max(sigma + bias,
                             torch.full(size=[], fill_value=epsilon, device=input.device, dtype=input.dtype))


def nnef_l2_normalization(input, axes, bias=0.0, epsilon=0.0):
    # type: (torch.Tensor, List[int], float, float)->torch.Tensor
    sigma = torch.sqrt(nnef_sum_reduce(torch.pow(input, 2.0), axes=axes))
    return input / torch.max(sigma + bias,
                             torch.full(size=[], fill_value=epsilon, device=input.device, dtype=input.dtype))


def nnef_matmul(A, B, transposeA=False, transposeB=False):
    # type:(torch.Tensor, torch.Tensor, bool, bool)->torch.Tensor

    return torch.matmul(torch.transpose(A, len(A.shape) - 2, len(A.shape) - 1) if transposeA else A,
                        torch.transpose(B, len(B.shape) - 2, len(B.shape) - 1) if transposeB else B)


def nnef_split(value, axis, ratios):
    # type:(torch.Tensor, int, List[int])->torch.Tensor
    assert value.shape[axis] % sum(ratios) == 0

    multiplier = value.shape[axis] // sum(ratios)
    sections = [ratio * multiplier for ratio in ratios]
    return torch.split(value, split_size_or_sections=sections, dim=axis)


def nnef_slice(input, axes, begin, end, stride=None):
    # type:(torch.Tensor, List[int], List[int], List[int], List[int])->torch.Tensor

    if stride is None:
        stride = [1] * len(axes)

    shape = list(input.shape)
    slices = [slice(None)] * len(shape)

    for axis, b, e, s in zip(axes, begin, end, stride):
        if b < 0:
            b += shape[axis]
        if e < 0:
            e += shape[axis]
        elif e == 0 and s == 1:
            e = shape[axis]

        b = _clamp(b, -1, shape[axis])
        e = _clamp(e, -1, shape[axis])

        if s > 0:
            slices[axis] = slice(b, e, s)
        else:
            offs = (b - e - 1) % (-s) + 1 if b != e else 1
            slices[axis] = slice(e+offs, b+1, -s)

    input = input[slices]

    flip_axes = [axis for axis, s in zip(axes, stride) if s < 0]
    if len(flip_axes) != 0:
        input = torch.flip(input, dims=flip_axes)

    return input


def nnef_select(condition, true_value, false_value):
    # type:(torch.Tensor, torch.Tensor, torch.Tensor)->torch.Tensor
    rank = max(len(condition.shape), len(true_value.shape), len(false_value.shape))
    return torch.where(_expand_to_rank(condition, rank),
                       _expand_to_rank(true_value, rank),
                       _expand_to_rank(false_value, rank))


def _nnef_generic_reduce(input, axes, f):
    # type:(torch.Tensor, List[int], Callable)->torch.Tensor
    if not axes:
        return input
    for axis in reversed(sorted(axes)):
        input = f(input=input, dim=axis, keepdim=True)
    return input


def nnef_sum_reduce(input, axes, normalize=False):
    # type:(torch.Tensor, List[int], bool)->torch.Tensor
    return _nnef_generic_reduce(input=input, axes=axes, f=torch.mean if normalize else torch.sum)


def nnef_max_reduce(input, axes):
    # type:(torch.Tensor, List[int])->torch.Tensor
    return _nnef_generic_reduce(input=input, axes=axes,
                                f=lambda input, dim, keepdim: torch.max(input, dim=dim, keepdim=keepdim)[0])


def nnef_min_reduce(input, axes):
    # type:(torch.Tensor, List[int])->torch.Tensor
    return _nnef_generic_reduce(input=input, axes=axes,
                                f=lambda input, dim, keepdim: torch.min(input, dim=dim, keepdim=keepdim)[0])


def nnef_mean_reduce(input, axes):
    # type:(torch.Tensor, List[int])->torch.Tensor
    return _nnef_generic_reduce(input=input, axes=axes, f=torch.mean)


def _nnef_argminmax_reduce(input, axes, argmin=False):
    # type:(torch.Tensor, List[int], bool)->torch.Tensor
    if len(axes) == 1:
        return _nnef_generic_reduce(input=input, axes=axes, f=torch.argmin if argmin else torch.argmax)
    else:
        axes = sorted(axes)
        consecutive_axes = list(range(axes[0], axes[0] + len(axes)))

        assert axes == consecutive_axes, \
            "{} is only implemented for consecutive axes.".format("argmin_reduce" if argmin else "argmax_reduce")

        reshaped = nnef_reshape(input,
                                shape=(list(input.shape)[:axes[0]]
                                       + [-1]
                                       + list(input.shape[axes[0] + len(axes):])))
        reduced = _nnef_generic_reduce(input=reshaped, axes=[axes[0]], f=torch.argmin if argmin else torch.argmax)
        reshaped = nnef_reshape(reduced, shape=list(dim if axis not in axes else 1
                                                    for axis, dim in enumerate(input.shape)))
        return reshaped


def nnef_argmax_reduce(input, axes):
    # type:(torch.Tensor, List[int])->torch.Tensor
    return _nnef_argminmax_reduce(input, axes, argmin=False)


def nnef_argmin_reduce(input, axes):
    # type:(torch.Tensor, List[int])->torch.Tensor
    return _nnef_argminmax_reduce(input, axes, argmin=True)


def nnef_clamp(x, a, b):
    # type:(torch.Tensor, torch.Tensor, torch.Tensor)->torch.Tensor
    rank = max(len(x.shape), len(a.shape), len(b.shape))
    x = _expand_to_rank(x, rank)
    a = _expand_to_rank(a, rank)
    b = _expand_to_rank(b, rank)
    return torch.max(torch.min(x, b), a)


def nnef_nearest_downsample(input, factor):
    # type: (torch.Tensor, List[int])->torch.Tensor
    dims = len(input.shape)
    return nnef_box(input, size=[1] * dims, stride=[1, 1] + factor, padding=[(0, 0)] * dims)


def nnef_area_downsample(input, factor):
    # type: (torch.Tensor, List[int])->torch.Tensor
    dims = len(input.shape)
    return nnef_box(input, size=[1, 1] + factor, stride=[1, 1] + factor, padding=[(0, 0)] * dims, normalize=True)


def nnef_moments(input, axes):
    # type: (torch.Tensor, List[int])->Tuple[torch.Tensor, torch.Tensor]
    mean = nnef_mean_reduce(input, axes=axes)
    variance = nnef_mean_reduce(torch.pow(input - mean, 2.0), axes=axes)
    return mean, variance


def nnef_linear(input, filter, bias):
    # type: (torch.Tensor, torch.Tensor, torch.Tensor)->torch.Tensor
    matmul = nnef_matmul(A=input, B=filter, transposeB=True)
    matmul, bias = _expand_binary(matmul, bias)
    return matmul + bias


def nnef_separable_conv(input,  # type: torch.Tensor
                        plane_filter,  # type: torch.Tensor
                        point_filter,  # type: torch.Tensor
                        bias,  # type: torch.Tensor
                        border='constant',  # type: str
                        padding=None,  # type: Optional[List[Tuple[int, int]]]
                        stride=None,  # type: Optional[List[int]]
                        dilation=None,  # type: Optional[List[int]]
                        groups=1,  # type: int
                        ):
    # type: (...)->torch.Tensor
    filtered = nnef_conv(input, plane_filter,
                         bias=torch.zeros(size=tuple(), device=input.device, dtype=input.dtype),
                         border=border,
                         padding=padding,
                         stride=stride,
                         dilation=dilation,
                         groups=0)
    return nnef_conv(filtered, point_filter, bias, groups=groups)


def nnef_separable_deconv(input,  # type: torch.Tensor
                          plane_filter,  # type: torch.Tensor
                          point_filter,  # type: torch.Tensor
                          bias,  # type: torch.Tensor
                          border='constant',  # type: str
                          padding=None,  # type: Optional[List[Tuple[int, int]]]
                          stride=None,  # type: Optional[List[int]]
                          dilation=None,  # type: Optional[List[int]]
                          output_shape=None,  # type: Optional[List[int]]
                          groups=1,  # type: int
                          ):
    # type: (...)->torch.Tensor
    filtered = nnef_deconv(input,
                           point_filter,
                           torch.zeros(size=tuple(), device=input.device, dtype=input.dtype),
                           groups=groups)
    return nnef_deconv(filtered, plane_filter, bias,
                       border=border,
                       padding=padding,
                       stride=stride,
                       dilation=dilation,
                       output_shape=output_shape,
                       groups=0)


def nnef_copy_n(x, times):
    # type: (torch.Tensor, int)->List[torch.Tensor]
    return [x.clone() for _ in range(times)]


def nnef_zero_point_linear_quantize(x, zero_point, scale, bits, signed, symmetric):
    # type: (torch.Tensor, torch.Tensor, torch.Tensor, int, bool, bool)->torch.Tensor

    z = torch.round(x / scale) + zero_point
    r = 2 ** (bits - 1) - 1 if signed else 2 ** bits - 1
    q = torch.clamp(z, 0 if not signed else -r if symmetric else -r - 1, r)
    y = (q - zero_point) * scale
    return y.type(x.dtype)


def nnef_min_max_linear_quantize(x, min, max, bits, signed, symmetric):
    # type: (torch.Tensor, torch.Tensor, torch.Tensor, int, bool, bool)->torch.Tensor

    r = float(2 ** bits - 1 - int(signed and symmetric))
    z = torch.clamp(x, min, max)
    q = torch.round((z - min) / (max - min) * r)
    return q * ((max - min) / r) + min


def nnef_logarithmic_quantize(x, max, bits):
    # type: (torch.Tensor, torch.Tensor, int)->torch.Tensor

    r = float(2 ** bits - 1)
    m = math.ceil(math.log2(max))
    q = torch.round(torch.clamp(torch.log2(torch.abs(x)), m - r, m))
    return torch.sign(x) * torch.pow(2.0, q)


def nnef_reshape(input, shape, axis_start=0, axis_count=-1):
    # type: (torch.Tensor, List[int], int, int)->torch.Tensor

    return input.reshape(nnef.shapes.reshape_shape(input=list(input.shape),
                                                   shape=shape,
                                                   axis_start=axis_start,
                                                   axis_count=axis_count))


def nnef_update(variable, value):
    # type: (torch.Tensor, torch.Tensor)->torch.Tensor
    return value


def nnef_transpose(input, axes):
    return input.permute(*(axes + list(range(len(axes), len(input.shape)))))


def nnef_squeeze(input, axes):
    return input.reshape(nnef.shapes.squeeze_shape(input.shape, axes))


def nnef_unsqueeze(input, axes):
    return input.reshape(nnef.shapes.unsqueeze_shape(input.shape, axes))


def nnef_cast(input, dtype):
    return input.to(_numpy_dtype_to_torch[dtype])


def nnef_gather(input, indices, axis):
    shape = tuple(indices.shape)
    if len(shape) != 1:
        indices = torch.flatten(indices)
    result = input.index_select(dim=axis, index=indices.to(torch.int64))
    if len(shape) != 1:
        result = torch.reshape(result, shape=input.shape[:axis] + shape + input.shape[axis + 1:])
    return result


"""
The supported operators
"""
Operators = {
    'update': nnef_update,
    'reshape': nnef_reshape,
    'transpose': nnef_transpose,
    'concat': lambda values, axis: torch.cat(values, axis),
    'split': nnef_split,
    'slice': nnef_slice,
    'squeeze': nnef_squeeze,
    'unsqueeze': nnef_unsqueeze,
    'stack': lambda values, axis: torch.stack(values, axis),
    'unstack': lambda value, axis: torch.unbind(value, axis),
    'add': nnef_add,
    'add_n': nnef_add_n,
    'sub': _binary(lambda x, y: x - y),
    'mul': _binary(lambda x, y: x * y),
    'div': _binary(lambda x, y: x / y),
    'pow': _binary(torch.pow),
    'exp': torch.exp,
    'log': torch.log,
    'abs': torch.abs,
    'sign': torch.sign,
    'rcp': torch.reciprocal,
    'neg': torch.neg,
    'copy': torch.clone,
    'lt': _binary(lambda x, y: x < y),
    'gt': _binary(lambda x, y: x > y),
    'le': _binary(lambda x, y: x <= y),
    'ge': _binary(lambda x, y: x >= y),
    'eq': _binary(torch.eq),
    'ne': _binary(torch.ne),
    'and': _binary(lambda x, y: x & y),
    'or': _binary(lambda x, y: x | y),
    'not': lambda x: ~x,
    'floor': torch.floor,
    'ceil': torch.ceil,
    'round': torch.round,
    'select': nnef_select,
    'sqr': lambda x: torch.pow(x, 2.0),
    'sqrt': torch.sqrt,
    'rsqr': lambda x: torch.pow(x, -2.0),
    'rsqrt': torch.rsqrt,
    'log2': torch.log2,
    'min': _binary(torch.min),
    'max': _binary(torch.max),
    'clamp': nnef_clamp,
    'matmul': nnef_matmul,
    'conv': nnef_conv,
    'deconv': nnef_deconv,
    'box': nnef_box,
    'debox': nnef_debox,
    'argmax_pool': nnef_argmax_pool,
    # 'sample': unsupported,
    'desample': nnef_desample,
    'nearest_downsample': nnef_nearest_downsample,
    'area_downsample': nnef_area_downsample,
    'nearest_upsample': nnef_nearest_upsample,
    'multilinear_upsample': nnef_multilinear_upsample,
    'sum_reduce': nnef_sum_reduce,
    'max_reduce': nnef_max_reduce,
    'min_reduce': nnef_min_reduce,
    'argmax_reduce': nnef_argmax_reduce,
    'argmin_reduce': nnef_argmin_reduce,
    'mean_reduce': nnef_mean_reduce,
    'moments': nnef_moments,
    'relu': F.relu,
    'sigmoid': torch.sigmoid,
    'softabs': lambda x, epsilon: torch.sqrt(torch.pow(x, 2.0) + epsilon),
    'softmax': nnef_softmax,
    'softplus': lambda x: torch.log(torch.exp(x) + 1.0),
    'elu': F.elu,
    'selu': lambda x, alpha, _lambda_: F.selu(x),
    'gelu': F.gelu,
    'silu': lambda x: x * torch.sigmoid(x),
    'prelu': lambda x, alpha: F.prelu(x, alpha),
    'leaky_relu': lambda x, alpha: F.leaky_relu(x, alpha),
    'max_pool_with_index': nnef_max_pool_with_index,
    'max_pool': nnef_max_pool,
    'avg_pool': nnef_avg_pool,
    'rms_pool': nnef_rms_pool,
    'linear': nnef_linear,
    'separable_conv': nnef_separable_conv,
    'separable_deconv': nnef_separable_deconv,
    'local_response_normalization': nnef_local_response_normalization,
    'local_mean_normalization': nnef_local_mean_normalization,
    'local_variance_normalization': nnef_local_variance_normalization,
    'local_contrast_normalization': nnef_local_contrast_normalization,
    'l1_normalization': nnef_l1_normalization,
    'l2_normalization': nnef_l2_normalization,
    'batch_normalization': nnef_batch_normalization,
    # 'avg_roi_pool': unsupported,
    # 'max_roi_pool': unsupported,
    # 'roi_resample': unsupported,
    # 'avg_roi_align': unsupported,
    # 'max_roi_align': unsupported,
    'linear_quantize': nnef_min_max_linear_quantize,
    'min_max_linear_quantize': nnef_min_max_linear_quantize,
    'zero_point_linear_quantize': nnef_zero_point_linear_quantize,
    'logarithmic_quantize': nnef_logarithmic_quantize,
    'copy_n': nnef_copy_n,
    'sin': lambda x: torch.sin(x),
    'cos': lambda x: torch.cos(x),
    'tan': lambda x: torch.tan(x),
    'asin': lambda x: torch.asin(x),
    'acos': lambda x: torch.acos(x),
    'atan': lambda x: torch.atan(x),
    'sinh': lambda x: torch.sinh(x),
    'cosh': lambda x: torch.cosh(x),
    'tanh': lambda x: torch.tanh(x),
    'asinh': lambda x: torch.asinh(x),
    'acosh': lambda x: torch.acosh(x),
    'atanh': lambda x: torch.atanh(x),
    'tile': lambda input, repeats: input.repeat(*repeats),
    'pad': nnef_pad,
    'cast': nnef_cast,
    'gather': nnef_gather,
    'any_reduce': lambda input, axes: _nnef_generic_reduce(input, axes=axes, f=torch.any),
    'all_reduce': lambda input, axes: _nnef_generic_reduce(input, axes=axes, f=torch.all),
}



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from .io.nnef import Reader
from .io.nnef.writer import _DtypeFromNumpy
from graphviz import Digraph
import numpy as np
import os


def _text_with_size(text, size):
    return '<FONT POINT-SIZE="{}">{}</FONT>'.format(size, text)


def _format_tensor_label(tensor):
    return '< {}<BR/> {}>'.format(tensor.name, _text_with_size(_dtype_str(tensor.dtype) + _shape_str(tensor.shape), size=10))


def _dtype_str(dtype):
    return _DtypeFromNumpy[dtype.type if isinstance(dtype, np.dtype) else dtype]


def _shape_str(shape):
    return '[' + ','.join(str(s) if s is not None else '?' for s in shape) + ']' if shape is not None else ''


def _truncate(text, max_length=32):
    return text[:max_length - 3] + "..." if len(text) > max_length else text


def _attribs_str(attribs, separator):
    s = []
    for k, v in sorted(attribs.items(), key=lambda e: e[0]):
        if k == "label":
            v = ".../" + v.split('/')[-1]
        elif k == "dtype":
            v = _dtype_str(v)
        s.append("{}{}{}".format(k, separator, _truncate(str(v))))
    return s


def _format_op_details(op):
    s = [
        "inputs: " + ", ".join(str(tensor.name if tensor.producer is not None else tensor.data) for tensor in op.inputs),
        "outputs: " + ", ".join(str(tensor.name) for tensor in op.outputs),
    ]

    s.extend(_attribs_str(op.attribs, separator=': '))

    return "&#13;&#10;".join(s)


def _format_op_label(op):
    attrs = (_text_with_size(s, size=10) for s in _attribs_str(op.attribs, separator='='))
    return '<{}<BR/>{}>'.format(op.type, '<BR/>'.join(attrs)) if len(op.attribs) else op.type


def _generate_digraph(graph, show_variables, verbose):
    digraph = Digraph()

    for op in graph.operations:
        if (show_variables or op.type != "variable") and op.type != "external":
            digraph.node(str(id(op)), _format_op_label(op) if verbose else op.type, shape="box", tooltip=_format_op_details(op))

    for tensor in graph.tensors:
        if tensor.producer is not None and (show_variables or tensor.producer.type != "variable") and tensor.producer.type != "external":
            for consumer in tensor.consumers:
                digraph.edge(str(id(tensor.producer)), str(id(consumer)),
                             label=_format_tensor_label(tensor) if verbose else "  " + tensor.name,
                             labeltooltip="{}{}".format(_dtype_str(tensor.dtype), _shape_str(tensor.shape)))

    for tensor in graph.inputs:
        digraph.node(str(id(tensor)), _format_tensor_label(tensor) if verbose else tensor.name, shape="ellipse",
                     tooltip="{}{}".format(_dtype_str(tensor.dtype), _shape_str(tensor.shape)))
        for consumer in tensor.consumers:
            digraph.edge(str(id(tensor)), str(id(consumer)), label=None)

    for tensor in graph.outputs:
        digraph.node(str(id(tensor)), _format_tensor_label(tensor) if verbose else tensor.name, shape="ellipse",
                     tooltip="{}{}".format(_dtype_str(tensor.dtype), _shape_str(tensor.shape)))
        digraph.edge(str(id(tensor.producer)), str(id(tensor)), label=None)

    return digraph


def main(args):
    reader = Reader(decomposed=args.decompose, infer_shapes=args.infer_shapes)
    graph = reader(args.model)
    digraph = _generate_digraph(graph, args.show_variables, args.verbose)
    digraph.render(args.model + '.gv', format=args.format, cleanup=True)
    os.rename(args.model + '.gv.' + args.format, args.model + '.' + args.format)
    return 0


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('model', type=str,
                        help='The model to visualize')
    parser.add_argument('--decompose', type=str, nargs='*', default=None,
                        help='Names of operators to be decomposed by NNEF parser')
    parser.add_argument('--verbose', action='store_true',
                        help='Add more info to the nodes and edges')
    parser.add_argument('--show-variables', action='store_true',
                        help='Show variables explicitly')
    parser.add_argument('--infer-shapes', action='store_true',
                        help='Perform shape inference and show in visualized graph')
    parser.add_argument('--format', type=str, choices=['svg', 'pdf', 'png', 'dot'], default='svg',
                        help='The format of the output')
    exit(main(parser.parse_args()))



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tflite

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class SubGraph(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsSubGraph(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = SubGraph()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def SubGraphBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x46\x4C\x33", size_prefixed=size_prefixed)

    # SubGraph
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # SubGraph
    def Tensors(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Tensor import Tensor
            obj = Tensor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # SubGraph
    def TensorsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SubGraph
    def TensorsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # SubGraph
    def Inputs(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # SubGraph
    def InputsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # SubGraph
    def InputsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SubGraph
    def InputsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # SubGraph
    def Outputs(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # SubGraph
    def OutputsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # SubGraph
    def OutputsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SubGraph
    def OutputsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # SubGraph
    def Operators(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Operator import Operator
            obj = Operator()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # SubGraph
    def OperatorsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SubGraph
    def OperatorsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # SubGraph
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def SubGraphStart(builder): builder.StartObject(5)
def SubGraphAddTensors(builder, tensors): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(tensors), 0)
def SubGraphStartTensorsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def SubGraphAddInputs(builder, inputs): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(inputs), 0)
def SubGraphStartInputsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def SubGraphAddOutputs(builder, outputs): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(outputs), 0)
def SubGraphStartOutputsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def SubGraphAddOperators(builder, operators): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(operators), 0)
def SubGraphStartOperatorsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def SubGraphAddName(builder, name): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)
def SubGraphEnd(builder): return builder.EndObject()



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division, print_function, absolute_import

from collections import namedtuple
from ....model import *
from ....utils.types import as_str
from .protobuf import *
import numpy as np
import six


Function = namedtuple('Function', ['name', 'attrs'])


_DtypeToNumpy = {
    'DT_INVALID': None,
    'DT_RESOURCE': np.dtype([('resource', np.int32)]),
    'DT_HALF': np.float16,
    'DT_FLOAT': np.float32,
    'DT_DOUBLE': np.float64,
    'DT_INT8': np.int8,
    'DT_INT16': np.int16,
    'DT_INT32': np.int32,
    'DT_INT64': np.int64,
    'DT_UINT8': np.uint8,
    'DT_UINT16': np.uint16,
    'DT_UINT32': np.uint32,
    'DT_UINT64': np.uint64,
    'DT_BOOL': np.bool_,
    'DT_STRING': np.str_,
    'DT_COMPLEX64': np.complex64,
    'DT_COMPLEX128': np.complex128,
}


def _get_shape(shape_proto):
    return tuple(int(dim.size) if dim.size >= 0 else None for dim in shape_proto.dim) \
        if not shape_proto.unknown_rank else None


def _get_dtype(dtype_enum):
    dtype = _DtypeToNumpy[DataType.Name(dtype_enum)]
    assert dtype is not None, "non-numeric dtype '{}' in attribute".format(DataType.Name(dtype_enum))
    return dtype


def _get_nonempty_items(message, fields):
    for field in fields:
        items = getattr(message, field)
        if len(items):
            return field, items

    return None, None


def _get_tensor(tensor_proto):
    shape = _get_shape(tensor_proto.tensor_shape)
    dtype = _get_dtype(tensor_proto.dtype)

    if len(tensor_proto.tensor_content):
        data = np.frombuffer(tensor_proto.tensor_content, dtype=dtype).reshape(shape)
    else:
        field, items = _get_nonempty_items(tensor_proto,
                                           fields=['half_val', 'float_val', 'double_val', 'int_val', 'int64_val',
                                                   'bool_val', 'string_val', 'uint32_val', 'uint64_val',
                                                   'resource_handle_val', 'scomplex_val', 'dcomplex_val'])

        if items is None and any(s == 0 for s in shape):
            items = []

        assert items is not None, "tensor items are empty, dtype = {}, shape = {}".format(dtype, shape)

        items = [item for item in items]
        if len(items) == int(np.prod(shape)):
            data = np.array(items, dtype=dtype).reshape(shape)
        else:
            assert len(items) == 1
            data = np.full(shape=shape, dtype=dtype, fill_value=items[0])

    return data


def _get_func(name_attrlist_proto):
    return Function(name_attrlist_proto.name, _get_attributes(name_attrlist_proto.attr))


def _get_attribute(field, value):
    if field == 'i' or field == 'f' or field == 'b' or field == 'placeholder':
        return value
    elif field == 's':
        return as_str(value.decode())
    elif field == 'shape':
        return _get_shape(value)
    elif field == 'type':
        return _get_dtype(value)
    elif field == 'tensor':
        return _get_tensor(value)
    elif field == 'func':
        return _get_func(value)
    elif field == 'list':
        field, items = _get_nonempty_items(value, fields=['i', 'f', 'b', 's', 'shape', 'type', 'tensor', 'func'])
        return [_get_attribute(field, item) for item in items] if items is not None else []

    assert False


def _get_attributes(attr_map_proto):
    attributes = {}
    for name, value in attr_map_proto.items():
        field = value.WhichOneof('value')
        if field is not None:
            value = getattr(value, field)
            attributes[as_str(name)] = _get_attribute(field, value)
        else:
            attributes[as_str(name)] = None

    return attributes


def _get_output_name(node_name, idx):
    return node_name + ':' + str(idx) if idx > 0 else node_name


def _has_output_shapes(graph_def):
    return all('_output_shapes' in node.attr and node.attr['_output_shapes'].WhichOneof('value') is not None
               for node in graph_def.node)


def _add_output_shapes(graph_def):
    try:
        import tensorflow.compat.v1 as tf
    except ImportError:
        import tensorflow as tf

    graph = tf.Graph()
    with graph.as_default():
        tf.import_graph_def(graph_def, name='')
        return graph.as_graph_def(add_shapes=True)


def _get_dtypes(graph_def):
    try:
        import tensorflow.compat.v1 as tf
    except ImportError:
        import tensorflow as tf

    dtypes = {}

    graph = tf.Graph()
    with graph.as_default():
        tf.import_graph_def(graph_def, name='')
        for op in graph.get_operations():
            for tensor in op.outputs:
                name = tensor.name[:-2] if tensor.name.endswith(':0') else tensor.name
                dtypes[name] = tensor.dtype.as_numpy_dtype if tensor.dtype != tf.resource else _DtypeToNumpy['DT_RESOURCE'].type

    return dtypes


def _get_output_shapes(attr_map_proto):
    value = attr_map_proto['_output_shapes']
    field = value.WhichOneof('value')
    if field is None:
        return None

    value = getattr(value, field)
    return _get_attribute(field, value)


def build_graph(graph_def):
    graph = Graph()

    dtypes = _get_dtypes(graph_def)

    # create tensors
    node_outputs = {}
    for node in graph_def.node:
        output_shapes = _get_output_shapes(node.attr)
        if output_shapes is not None:
            name = as_str(node.name)
            node_outputs[name] = [Tensor(graph, _get_output_name(name, idx), shape=shape, dtype=dtypes.get(name))
                                  for idx, shape in enumerate(output_shapes)]

    tensors = {tensor.name: tensor for outputs in six.itervalues(node_outputs) for tensor in outputs}

    # create ops
    for node in graph_def.node:
        attributes = _get_attributes(node.attr)
        inputs = [tensors[name] for name in node.input if not name.startswith('^')]
        outputs = node_outputs[node.name] if node.name in node_outputs else []

        Operation(graph,
                  type=as_str(node.op),
                  name=as_str(node.name),
                  inputs=inputs,
                  outputs=outputs,
                  attribs=attributes)

    graph.inputs = [node_outputs[node.name][0] for node in graph_def.node if node.op == 'Placeholder']
    graph.outputs = [output for op in graph.operations if all(len(output.consumers) == 0 for output in op.outputs)
                     for output in op.outputs]
    return graph


def _unpack_custom_ops(graph):
    for op in graph.operations:
        if op.type == 'PyFunc':
            op.custom = True
            op.type = op.attribs['token']
            op.attribs = {key[2:-2]: value for key, value in six.iteritems(op.attribs)
                          if key.startswith('_$') and key.endswith('$_')}


def read_graphdef(filename, input_shapes, fold_constants):
    graph_def = GraphDef()
    with open(filename, 'rb') as file:
        graph_def.ParseFromString(file.read())

    if not _has_output_shapes(graph_def):
        graph_def = _add_output_shapes(graph_def)

    if input_shapes is not None:
        from .utils import set_input_shapes
        graph_def = set_input_shapes(graph_def, input_shapes)

    if fold_constants:
        from .utils import fold_constant_tensors
        graph_def = fold_constant_tensors(graph_def)

    graph = build_graph(graph_def)
    _unpack_custom_ops(graph)

    return graph


class Reader(object):

    def __init__(self, fold_constants=False):
        self._fold_constants = fold_constants

    def __call__(self, filename, input_shapes=None):
        return read_graphdef(filename, input_shapes, self._fold_constants)



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division, print_function, absolute_import
from .converter import ConverterFromNNEF as _Converter, Transform, ConversionError
from .nnef_to_tf import Converter as _TFConverter, _Transforms as _TFTransforms
from ..model import Tensor, Operation
from ..model.utils import generate_tensor_names_from_op_type
from ..io.tf.lite import CustomOptionsKey
import numpy as np
import copy


def tflite_detection_postprocess_shape(input, scores, anchors, **kwargs):
    return [], [], [], []


class Converter(_TFConverter):

    @staticmethod
    def defined_shapes():
        return {
            'relu6': lambda shape: shape,
            'TFLite_Detection_PostProcess': tflite_detection_postprocess_shape,
        }

    @staticmethod
    def decomposed_operations():
        return _Converter.decomposed_operations()

    def __init__(self, io_transpose=False, custom_transforms=None, custom_functions=None, mirror_unsupported=False):
        _Converter.__init__(self, transforms=self.merge_transforms(_Transforms, custom_transforms),
                            functions=custom_functions, mirror_unsupported=mirror_unsupported)
        self._data_format = 'NXC'
        self._io_transpose = io_transpose

    def __call__(self, graph):
        graph = _TFConverter.__call__(self, graph)
        self._generate_tensor_names(graph)
        self._fix_custom_options(graph)
        return graph

    def _global_attribs(self):
        return {'_lite_': True}

    def _prepare(self, graph):
        self._fix_quantized_dtypes(graph)
        self._fix_quantization_attribs(graph)
        self._transpose_externals(graph)

    def _transpose_externals(self, graph):
        for tensor in graph.tensors:
            mapped = self._tensor_map[tensor]
            if mapped.producer and mapped.producer.type == 'external' and self.needs_io_transpose(tensor):
                self._transposes[tensor] = self.ncx_to_nxc(tensor.shape)

    def _generate_tensor_names(self, graph):
        generate_tensor_names_from_op_type(graph)

        placeholders = 0
        constants = 0
        for tensor in graph.tensors:
            if tensor.name is None:
                if tensor.data is None:
                    placeholders += 1
                    tensor.name = 'PLACEHOLDER' + str(placeholders)
                else:
                    constants += 1
                    tensor.name = 'CONSTANT' + str(constants)

    def _fix_quantized_dtypes(self, graph):
        for tensor in graph.tensors:
            if tensor.quant and tensor.dtype == np.float32:
                bits = tensor.quant['bits']
                signed = tensor.quant['signed']
                assert bits == 8 or bits == 32
                tensor.dtype = (np.int8 if signed else np.uint8) if bits == 8 else (np.int32 if signed else np.uint32)

    def _fix_quantization_attribs(self, graph):
        for tensor in graph.tensors:
            if tensor.quant:
                opname = tensor.quant['op-name']
                if opname != 'zero_point_linear_quantize':
                    raise ConversionError("Quantization operation '{}' cannot be converted to TFLite")

                del tensor.quant['op-name']
                del tensor.quant['bits']
                if 'signed' in tensor.quant:
                    del tensor.quant['signed']
                if 'symmetric' in tensor.quant:
                    del tensor.quant['symmetric']

    def _fix_custom_options(self, graph):
        for op in graph.operations:
            if op.custom:
                options = op.attribs.get(CustomOptionsKey)
                if options is not None:
                    op.attribs[CustomOptionsKey] = bytes.fromhex(options)

    def _make_constant(self, graph, dtype, value, inline):
        return Tensor(graph, dtype=dtype, shape=self._shape_of(value), data=value)

    def _ensure_constant_producer(self, tensor):
        pass

    def _transform_constant(self, tensor, func):
        data = func(tensor.data)
        tensor.shape = data.shape
        tensor.data = data

    def _squeeze_operation(self, input, output, axes):
        Operation(input.graph, type='SQUEEZE', inputs=input, outputs=output, attribs={'squeeze_dims': axes})

    def _unsqueeze_operation(self, input, output, axes):
        if len(axes) == 1:
            Operation(input.graph, type='EXPAND_DIMS', inputs=(input, self.as_tensor(axes[0], np.int32)),
                      outputs=output)
        else:
            Operation(input.graph, type='RESHAPE', inputs=(input, self.as_tensor(output.shape, np.int32)),
                      outputs=output, attribs={'new_shape': output.shape})

    def _transpose_operation(self, input, output, perm):
        Operation(input.graph, type='TRANSPOSE', inputs=(input, self.as_tensor(perm, np.int32)),
                  outputs=output)

    def _reshape_operation(self, input, output, shape):
        Operation(input.graph, type='RESHAPE', inputs=(input, self.as_tensor(shape, np.int32)), outputs=output,
                  attribs={'new_shape': shape})

    def _bias_operation(self, input, output, bias):
        if not isinstance(bias, Tensor):
            bias = self.as_tensor(bias, np.float32)

        Operation(input.graph, type='ADD', inputs=(input, bias), outputs=output)

    def _scale_operation(self, input, output, scalar):
        if not isinstance(scalar, Tensor):
            scalar = self.as_tensor(scalar, np.float32)

        Operation(input.graph, type='MUL', inputs=(input, scalar), outputs=output)

    def _pad_operation(self, input, output, paddings):
        if not isinstance(paddings, Tensor):
            paddings = self.as_tensor(paddings, np.int64)

        Operation(input.graph, type='PAD', inputs=(input, paddings), outputs=output, attribs={})

    def is_same_padding(self, input_size, output_size, stride):
        return all(o == i // s for i, o, s in zip(input_size, output_size, stride))

    def is_valid_padding(self, padding):
        return len(padding) != 0 and all(p == (0, 0) for p in padding)

    def pad_input(self, input, paddings):
        if all(item == (0, 0) for item in paddings):
            return input

        shape = tuple(p + x + q for x, (p, q) in zip(self._working_shape(input), paddings))
        output = Tensor(input.graph, dtype=input.dtype, shape=shape, quant=copy.deepcopy(input.quant))
        self._pad_operation(input, output, paddings)
        return output


_Transforms = Converter.unpack_transforms({
    ('external', 'constant'):
        Transform(type=None),
    'conv':
        Transform(
            type='!"CONV_2D" if not depthwise else "DEPTHWISE_CONV_2D"',
            cond={
                '!I[0].rank == 4': 'rank must be 4',
            },
            using={
                'depthwise': '!groups == 0',
                'channels': '!I[0].shape[1]',
                'valid_pad': '!is_valid_padding(padding)',
                'same_pad': '!is_same_padding(I[0].shape[2:], O[0].shape[2:], stride)',
                'pads': '![(0, 0)] + padding + [(0, 0)]',
            },
            inputs=(
                '!transpose_input(I[0]) if same_pad or valid_pad else pad_input(transpose_input(I[0]), pads)',
                '!transpose_filter(I[1], format="NXC" if not depthwise else "CXN")',
                '!squeeze_vector(I[2])',
            ),
            outputs='!transpose_output(O[0])',
            attribs={
                'stride_h': '!stride[0]',
                'stride_w': '!stride[1]',
                'dilation_h_factor': '!dilation[0]',
                'dilation_w_factor': '!dilation[1]',
                'padding': '!"VALID" if valid_pad else "SAME"',
                'depth_multiplier': '!O[0].shape[1] // channels if depthwise else None',
            }
        ),
    'deconv':
        Transform(
            type='TRANSPOSE_CONV',
            cond={
                '!I[0].rank == 4': 'rank must be 4',
                '!groups == 1': 'groups must be 1',
            },
            using={
                'depthwise': '!groups == 0',
                'channels': '!O[0].shape[1]',
                'valid_pad': '!is_valid_padding(padding)',
                'same_pad': '!is_same_padding(I[0].shape[2:], O[0].shape[2:], stride)',
                'pads': '![(0, 0)] + padding + [(0, 0)]',
            },
            inputs=(
                '!as_tensor(ncx_to_nxc(output_shape), np.int32)',
                '!transpose_filter(I[1], format="CXN" if not depthwise else "NXC")',
                '!transpose_input(I[0]) if same_pad or valid_pad else pad_input(transpose_input(I[0]), pads)',
            ),
            outputs='!bias_add(transpose_output(O[0]), squeeze_vector(I[2]) if I[2].rank == 2 else I[2])',
            attribs={
                'stride_h': '!stride[0]',
                'stride_w': '!stride[1]',
                'padding': '!"VALID" if valid_pad else "SAME"',
                'depth_multiplier': '!I[1].shape[0] // channels if depthwise else None',
            }
        ),
    ('max_pool', 'avg_pool'):
        Transform(
            cond={
                '!size[0] == 1 and size[1] == 1 and ': 'size must be 1 in batch and channel dimensions',
                '!stride[0] == 1 and stride[1] == 1': 'stride must be 1 in batch and channel dimensions',
            },
            type=('MAX_POOL_2D', 'AVERAGE_POOL_2D'),
            using={
                'valid_pad': '!is_valid_padding(padding)',
                'same_pad': '!is_same_padding(I[0].shape[2:], O[0].shape[2:], stride[2:])',
            },
            inputs=(
                '!transpose_input(I[0]) if same_pad or valid_pad else pad_input(transpose_input(I[0]), padding)',
            ),
            outputs=(
                '!transpose_output(O[0])',
            ),
            attribs={
                'filter_height': '!size[2]',
                'filter_width': '!size[3]',
                'stride_h': '!stride[2]',
                'stride_w': '!stride[3]',
                'padding': '!"VALID" if valid_pad else "SAME"',
            }
        ),
    'reshape':
        Transform(
            type='RESHAPE',
            using={
                'new_shape': '!fixed_batch(shape, I[0].shape[0])',
            },
            inputs=(
                '!undo_transpose(I[0])',
                '!as_tensor(new_shape, np.int32)',
            ),
            outputs='!O[0]',
            attribs={
                'new_shape': '!new_shape',
            }
        ),
    'concat':
        Transform(
            type='CONCATENATION',
            inputs=['!I[:]'],
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'axis': '!transpose_axis_like(axis, I[0])',
            }
        ),
    'copy':
        Transform(
            type='RESHAPE',
            using={
                'shape': '!transpose_list_like(I[0].shape, I[0])',
            },
            inputs=(
                '!I[0]',
                '!as_tensor(shape, np.int32)',
            ),
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'new_shape': '!shape',
            }
        ),
    'linear':
        Transform(
            type='FULLY_CONNECTED',
            inputs=(
                '!I[0]',
                '!I[1]',
                '!squeeze_vector(I[2]) if not is_zero(I[2]) else None',
            ),
            outputs='!O[0]',
            attribs={
                'fused_activation_function': "NONE",
                'weights_format': "DEFAULT",
                'keep_num_dims': True,
                'asymmetric_quantize_inputs': False,
            }
        ),
    'matmul':
        Transform(
            type='BATCH_MATMUL',
            inputs=(
                '!I[0]',
                '!I[1]',
            ),
            outputs='!O[0]',
            attribs={
                'adj_x': '!transposeA',
                'adj_y': '!transposeB',
                'asymmetric_quantize_inputs': False,
            }
        ),
    'batch_normalization':
        Transform(
            type='MUL',
            cond={
                '!I[1].data is not None and I[2].data is not None and'
                ' (len(I) == 3 or I[3].data is not None) and (len(I) == 4 or I[4].data is not None)':
                    'all parameters must be constants',
                '!not any(t.quant for t in I)': 'quantized inputs or parameters are not supported',
            },
            using={
                'mean': '!np.squeeze(I[1].data, axis=0) if I[1].data is not None else None',
                'std': '!np.squeeze(np.sqrt(I[2].data + epsilon), axis=0) if I[2].data is not None else None',
                'offset': '!np.squeeze(I[3].data, axis=0) if I[3].data is not None else None if len(I) > 3 else 0',
                'scale': '!np.squeeze(I[4].data, axis=0) if I[4].data is not None else None if len(I) > 4 else 1',
            },
            inputs=(
                '!transpose_input(I[0])',
                '!as_tensor(scale / std, np.float32)',
            ),
            outputs='!bias_add(transpose_like(O[0], I[0]), as_tensor(offset - scale * mean / std, np.float32))',
        ),
    'l2_normalization':
        Transform(
            type='L2_NORMALIZATION',
            cond={
                '!axes == list(range(I[0].rank))': 'axes must denote all dimensions',
            },
            inputs='!I[0]',
            outputs='!transpose_like(O[0], I[0])',
        ),
    'prelu':
        Transform(
            type='PRELU',
            inputs=('!I[0]', '!I[1]'),
            outputs='!transpose_like(O[0], I[0])',
        ),
    'pad':
        Transform(
            type='!"PAD" if border == "constant" else "MIRROR_PAD"',
            cond={
                '!border in ["constant", "reflect", "reflect-even"]':
                    'border must be one of "constant", "reflect", "reflect-even"',
            },
            using={'paddings': '![list(item) for item in padding]'},
            inputs=(
                '!I[0]',
                '!as_tensor(ncx_to_nxc(paddings, cond=transposing(I[0])), np.int32)',
            ),
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'mode': '!0 if border == "reflect" else 1 if border == "reflect-even" else None',
            },
        ),
    'gather':
        Transform(
            type='GATHER',
            inputs=('!I[0]', '!I[1]'),
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'axis': '!transpose_axis_like(axis, I[0])',
            },
        ),
    'cast':
        Transform(
            type='CAST',
            inputs='!I[0]',
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'in_data_type': '!I[0].dtype',
                'out_data_type': '!O[0].dtype',
            },
        ),
    # 'copy': _TFTransforms['copy'].with_type('IDENTITY'),  # only works in TF 2.3
    'transpose': _TFTransforms['transpose'].with_type('TRANSPOSE'),
    'split': _TFTransforms['split'].with_type('SPLIT_V'),
    'squeeze': _TFTransforms['squeeze'].with_type('SQUEEZE'),
    'unsqueeze': _TFTransforms['unsqueeze'].with_type('!"EXPAND_DIMS" if len(axes) == 1 else "RESHAPE"'),
    'relu': _TFTransforms['relu'].with_type('RELU'),
    'relu6': _TFTransforms['relu6'].with_type('RELU6'),
    'elu': _TFTransforms['elu'].with_type('ELU'),
    'leaky_relu': _TFTransforms['leaky_relu'].with_type('LEAKY_RELU'),
    'sigmoid': _TFTransforms['sigmoid'].with_type('LOGISTIC'),
    'sin': _TFTransforms['sin'].with_type('SIN'),
    'cos': _TFTransforms['cos'].with_type('COS'),
    'tan': _TFTransforms['tan'].with_type('TAN'),
    'asin': _TFTransforms['asin'].with_type('ASIN'),
    'acos': _TFTransforms['acos'].with_type('ACOS'),
    'atan': _TFTransforms['atan'].with_type('ATAN'),
    'sinh': _TFTransforms['sinh'].with_type('SINH'),
    'cosh': _TFTransforms['cosh'].with_type('COSH'),
    'tanh': _TFTransforms['tanh'].with_type('TANH'),
    'asinh': _TFTransforms['asinh'].with_type('ASINH'),
    'acosh': _TFTransforms['acosh'].with_type('ACOSH'),
    'atanh': _TFTransforms['atanh'].with_type('ATANH'),
    'exp': _TFTransforms['exp'].with_type('EXP'),
    'log': _TFTransforms['log'].with_type('LOG'),
    'abs': _TFTransforms['abs'].with_type('ABS'),
    'neg': _TFTransforms['neg'].with_type('NEG'),
    'not': _TFTransforms['not'].with_type('LOGICAL_NOT'),
    'floor': _TFTransforms['floor'].with_type('FLOOR'),
    'ceil': _TFTransforms['ceil'].with_type('CEIL'),
    'round': _TFTransforms['round'].with_type('ROUND'),
    'sqr': _TFTransforms['sqr'].with_type('SQUARE'),
    'sqrt': _TFTransforms['sqrt'].with_type('SQRT'),
    'rsqrt': _TFTransforms['rsqrt'].with_type('RSQRT'),
    'add': _TFTransforms['add'].with_type('ADD'),
    'sub': _TFTransforms['sub'].with_type('SUB'),
    'mul': _TFTransforms['mul'].with_type('MUL'),
    'div': _TFTransforms['div'].with_type('DIV'),
    'pow': _TFTransforms['pow'].with_type('POW'),
    'min': _TFTransforms['min'].with_type('MINIMUM'),
    'max': _TFTransforms['max'].with_type('MAXIMUM'),
    'and': _TFTransforms['and'].with_type('LOGICAL_AND'),
    'or': _TFTransforms['or'].with_type('LOGICAL_OR'),
    'lt': _TFTransforms['lt'].with_type('LESS'),
    'le': _TFTransforms['le'].with_type('LESS_EQUAL'),
    'gt': _TFTransforms['gt'].with_type('GREATER'),
    'ge': _TFTransforms['ge'].with_type('GREATER_EQUAL'),
    'eq': _TFTransforms['eq'].with_type('EQUAL'),
    'ne': _TFTransforms['ne'].with_type('NOT_EQUAL'),
    'select': _TFTransforms['select'].with_type('SELECT'),
    'min_reduce': _TFTransforms['min_reduce'].with_type('REDUCE_MIN'),
    'max_reduce': _TFTransforms['max_reduce'].with_type('REDUCE_MAX'),
    'mean_reduce': _TFTransforms['mean_reduce'].with_type('MEAN'),
    'sum_reduce': _TFTransforms['sum_reduce'].with_type('SUM'),
    'any_reduce': _TFTransforms['any_reduce'].with_type('REDUCE_ANY'),
    'all_reduce': _TFTransforms['all_reduce'].with_type('REDUCE_ALL'),
    'argmin_reduce': _TFTransforms['argmin_reduce'].with_type('ARG_MIN'),
    'argmax_reduce': _TFTransforms['argmax_reduce'].with_type('ARG_MAX'),
    'stack': _TFTransforms['stack'].with_type('PACK'),
    'unstack': _TFTransforms['unstack'].with_type('UNPACK'),
    'tile': _TFTransforms['tile'].with_type('TILE'),
    'slice': _TFTransforms['slice'].with_type('STRIDED_SLICE'),
    'softmax': _TFTransforms['softmax'].with_type('SOFTMAX'),
    'local_response_normalization': _TFTransforms['local_response_normalization'].with_type('LOCAL_RESPONSE_NORMALIZATION'),
    'nearest_upsample': _TFTransforms['nearest_upsample'].with_type('RESIZE_NEAREST_NEIGHBOR'),
    'nearest_downsample': _TFTransforms['nearest_downsample'].with_type('RESIZE_NEAREST_NEIGHBOR'),
    'multilinear_upsample': _TFTransforms['multilinear_upsample'].with_type('RESIZE_BILINEAR'),
    'add_n': _TFTransforms['add_n'].with_type('ADD_N'),
})



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tflite

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class SVDFOptions(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsSVDFOptions(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = SVDFOptions()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def SVDFOptionsBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x46\x4C\x33", size_prefixed=size_prefixed)

    # SVDFOptions
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # SVDFOptions
    def Rank(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # SVDFOptions
    def FusedActivationFunction(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # SVDFOptions
    def AsymmetricQuantizeInputs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

def SVDFOptionsStart(builder): builder.StartObject(3)
def SVDFOptionsAddRank(builder, rank): builder.PrependInt32Slot(0, rank, 0)
def SVDFOptionsAddFusedActivationFunction(builder, fusedActivationFunction): builder.PrependInt8Slot(1, fusedActivationFunction, 0)
def SVDFOptionsAddAsymmetricQuantizeInputs(builder, asymmetricQuantizeInputs): builder.PrependBoolSlot(2, asymmetricQuantizeInputs, 0)
def SVDFOptionsEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tflite

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Metadata(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsMetadata(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Metadata()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def MetadataBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x46\x4C\x33", size_prefixed=size_prefixed)

    # Metadata
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Metadata
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Metadata
    def Buffer(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

def MetadataStart(builder): builder.StartObject(2)
def MetadataAddName(builder, name): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)
def MetadataAddBuffer(builder, buffer): builder.PrependUint32Slot(1, buffer, 0)
def MetadataEnd(builder): return builder.EndObject()



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division, print_function, absolute_import

from .protobuf import *
import numpy as np
import six


_DtypeFromNumpy = {
    None: 'DT_INVALID',
    np.float16: 'DT_HALF',
    np.float32: 'DT_FLOAT',
    np.float64: 'DT_DOUBLE',
    np.int8: 'DT_INT8',
    np.int16: 'DT_INT16',
    np.int32: 'DT_INT32',
    np.int64: 'DT_INT64',
    np.uint8: 'DT_UINT8',
    np.uint16: 'DT_UINT16',
    np.uint32: 'DT_UINT32',
    np.uint64: 'DT_UINT64',
    np.bool_: 'DT_BOOL',
    np.str_: 'DT_STRING',
    np.complex64: 'DT_COMPLEX64',
    np.complex128: 'DT_COMPLEX128',
    np.dtype([('resource', np.int32)]): 'DT_RESOURCE',
}

_NumpyDtypes = {
    np.int8, np.int16, np.int32, np.int64,
    np.uint8, np.uint16, np.uint32, np.uint64,
    np.float32, np.float64,
    np.complex64, np.complex128,
    np.bool_, np.str_,
    np.dtype([('resource', np.int32)]),
}


def _build_shape(shape_proto, shape):
    shape_proto.unknown_rank = (shape is None)
    if shape is not None:
        for item in shape:
            dim = shape_proto.dim.add()
            dim.size = item if item is not None else -1


def _build_dtype(dtype):
    return DataType.Value(_DtypeFromNumpy[dtype])


def _build_tensor(tensor_proto, data):
    if data.dtype is not None:
        tensor_proto.dtype = _build_dtype(data.dtype.type)

    if data.shape is not None:
        _build_shape(tensor_proto.tensor_shape, data.shape)

    tensor_proto.tensor_content = data.reshape([-1]).view(np.uint8).tobytes()
    return tensor_proto


def _build_attribute(attr_proto, value):
    if value is None:
        return attr_proto

    if type(value) in _NumpyDtypes:
        value = np.array(value)

    if isinstance(value, bool):  # must be before int
        attr_proto.b = value
    elif isinstance(value, int):
        attr_proto.i = value
    elif isinstance(value, float):
        attr_proto.f = value
    elif isinstance(value, str):
        attr_proto.s = value.encode()
    elif isinstance(value, (type, np.dtype)):
        attr_proto.type = _build_dtype(value)
    elif isinstance(value, tuple):
        _build_shape(attr_proto.shape, value)
    elif isinstance(value, np.ndarray):
        _build_tensor(attr_proto.tensor, value)
    elif isinstance(value, list):
        if len(value) == 0:
            attr_proto.list.i.extend([])     # to signal that the 'list' is the active in the oneof field
        else:
            first = value[0]
            if isinstance(first, int):
                attr_proto.list.i.extend(value)
            elif isinstance(first, float):
                attr_proto.list.f.extend(value)
            elif isinstance(first, bool):
                attr_proto.list.b.extend(value)
            elif isinstance(first, str):
                attr_proto.list.s.extend([item.encode() for item in value])
            elif isinstance(first, (type, np.dtype)):
                attr_proto.list.type.extend([_build_dtype(item) for item in value])
            elif isinstance(first, tuple):
                for item in value:
                    _build_shape(attr_proto.list.shape.add(), item)
            elif isinstance(first, np.ndarray):
                for item in value:
                    _build_tensor(attr_proto.list.tensor.add(), item)
            else:
                raise TypeError('unable to build attribute proto message from type: ' + str(type(first)))
    else:
        raise TypeError('unable to build attribute proto message from type: ' + str(type(value)))

    return attr_proto


def _build_output_shapes(attr_proto, output_shapes):
    for item in output_shapes:
        _build_shape(attr_proto.list.shape.add(), item)


def _tensor_name(tensor):
    name = tensor.producer.name
    idx = tensor.producer.outputs.index(tensor)
    return name + ':' + str(idx) if idx > 0 else name


def _custom_attribs(operation):
    attribs = {'_$' + key + '$_': value for key, value in six.iteritems(operation.attribs)}
    attribs['token'] = operation.type
    attribs['Tin'] = [tensor.dtype for tensor in operation.inputs]
    attribs['Tout'] = [tensor.dtype for tensor in operation.outputs]


def _build_node(node_def, operation):
    node_def.op = operation.type if not operation.custom else 'PyFunc'
    node_def.name = operation.name
    node_def.input.extend([_tensor_name(tensor) for tensor in operation.inputs])

    attribs = operation.attribs if not operation.custom else _custom_attribs(operation)

    output_shapes = attribs.get('_output_shapes')
    if output_shapes is not None:
        _build_output_shapes(node_def.attr['_output_shapes'], output_shapes)
        del attribs['_output_shapes']
    else:
        _build_output_shapes(node_def.attr['_output_shapes'],
                             [tensor.shape for tensor in operation.outputs])

    for name, value in attribs.items():
        _build_attribute(node_def.attr[name], value)

    return node_def


def build_graphdef(graph):
    graph_def = GraphDef()
    for operation in graph.operations:
        node_def = graph_def.node.add()
        _build_node(node_def, operation)
    return graph_def


def write_graphdef(graph, filename):
    graph_def = build_graphdef(graph)

    with open(filename, 'wb') as file:
        file.write(graph_def.SerializeToString())


class Writer(object):

    def __call__(self, graph, filename):
        return write_graphdef(graph, filename)



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..model.utils import replace_chain
import six


class Optimizer:

    def __init__(self, custom_optimizers=None):
        self._custom_optimizers = custom_optimizers or {}

    def __call__(self, graph, only_required=False):
        self._fix_batchnorm_spatial(graph)
        for chain, replacer in six.iteritems(self._custom_optimizers):
            replace_chain(graph, chain, replacer)
        return graph

    @staticmethod
    def _fix_batchnorm_spatial(graph):
        for op in graph.operations:
            if op.type == 'BatchNormalization':
                spatial = op.attribs.get('spatial')
                if spatial == 0 and op.inputs[1].rank == 1:
                    del op.attribs['spatial']



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division, print_function, absolute_import

import os
import shutil
import tempfile
from collections import OrderedDict

import nnef
import numpy as np
import six

from ...model import *
from ...utils import types
from .helpers import tgz_extract


_DtypeToNumpy = {
    'scalar': np.float32,
    'integer': np.int64,
    'logical': np.bool_,
}


def _recursive_itemize(arg):
    if type(arg) is tuple or type(arg) is list:
        for item in arg:
            yield from _recursive_itemize(item)
    elif type(arg) is dict or type(arg) is OrderedDict:
        for item in six.itervalues(arg):
            yield from _recursive_itemize(item)
    else:
        yield arg


def _make_constant_tensor(graph, value):
    value = types.to_numpy(value)
    return Tensor(graph=graph, shape=(), dtype=value.dtype.type, data=value)


def _make_tensor(graph, nnef_tensor):
    dtype = nnef_tensor.data.dtype.type if isinstance(nnef_tensor.data, np.ndarray) else _DtypeToNumpy[nnef_tensor.dtype]
    return Tensor(graph=graph, name=nnef_tensor.name, shape=tuple(nnef_tensor.shape) if nnef_tensor.shape is not None else None,
                  dtype=dtype, data=nnef_tensor.data, quant=nnef_tensor.quantization)


def _build_graph(nnef_graph):
    graph = Graph(name=nnef_graph.name)

    tensor_by_name = {}
    for nnef_op in nnef_graph.operations:
        inputs = (tensor_by_name[item] if isinstance(item, nnef.Identifier) else _make_constant_tensor(graph, item)
                  for item in _recursive_itemize(nnef_op.inputs))
        inputs = list(inputs) if any(isinstance(item, list) for item in six.itervalues(nnef_op.inputs)) else tuple(inputs)

        outputs = (_make_tensor(graph, nnef_graph.tensors[str(item)])
                   for item in _recursive_itemize(nnef_op.outputs))
        outputs = list(outputs) if any(isinstance(item, list) for item in six.itervalues(nnef_op.outputs)) else tuple(outputs)

        for tensor in outputs:
            tensor_by_name[str(tensor.name)] = tensor

        attribs = dict(nnef_op.attribs)
        if nnef_op.dtype is not None:
            attribs['dtype'] = outputs[0].dtype if nnef_op.name == 'constant' or nnef_op.name == 'variable' else \
                _DtypeToNumpy[nnef_op.dtype]

        _substitute_empty_array(nnef_op.name, 'stride', attribs, inputs)
        _substitute_empty_array(nnef_op.name, 'dilation', attribs, inputs)

        custom = nnef_op.name not in nnef.StandardOperations

        Operation(graph=graph, type=nnef_op.name, attribs=attribs, inputs=inputs, outputs=outputs, custom=custom)

    graph.inputs = [tensor_by_name[str(item)] for item in nnef_graph.inputs]
    graph.outputs = [tensor_by_name[str(item)] for item in nnef_graph.outputs]

    return graph


def _substitute_empty_array(op, key, attribs, inputs):
    value = attribs.get(key)
    if value is not None and len(value) == 0:
        rank = None
        if op == 'slice':
            rank = len(attribs['axes'])
        elif len(inputs) > 0 and inputs[0].rank is not None:
            rank = inputs[0].rank - 2 if op.endswith('conv') else inputs[0].rank

        if rank is not None:
            attribs[key] = [1] * rank


class Reader(object):

    def __init__(self, stdlib=None, decomposed=None, custom_shapes=None, infer_shapes=True, load_variables=True):
        self._stdlib = stdlib
        self._decomposed = decomposed
        self._custom_shapes = custom_shapes
        self._infer_shapes = infer_shapes
        self._load_variables = load_variables

    def __call__(self, path, input_shapes=None):
        compressed = os.path.splitext(path) in ['tgz', 'gz'] and not os.path.isdir(path)

        folder = None
        try:
            if compressed:
                folder = tempfile.mkdtemp(prefix="nnef_")
                tgz_extract(path, folder)
                path = folder

            if not os.path.isdir(path):
                raise IOError("NNEF model must be a (compressed) folder, but an uncompressed file was provided")

            nnef_graph = nnef.load_graph(path, stdlib=self._stdlib, lowered=self._decomposed, load_variables=self._load_variables)
            if self._infer_shapes:
                nnef.infer_shapes(nnef_graph, external_shapes=input_shapes or {}, custom_shapes=self._custom_shapes or {})

            return _build_graph(nnef_graph)
        finally:
            if folder is not None:
                shutil.rmtree(folder)



from src.nnef_tools.io.nnef.reader import Reader
from src.nnef_tools.io.nnef.writer import Writer
import numpy as np
import argparse
import json
import os


_CONV_OPS = ['conv', 'deconv', 'separable_conv', 'separable_deconv']


def make_quantization(min, max, signed, symmetric):
    if min > 0:
        min = 0
    if max < 0:
        max = 0

    if signed and symmetric:
        if -max < min:
            min = -max
        if -min > max:
            max = -min

    scale = 255 / (max - min)
    zero_point = int((0 - min) * scale)

    if signed:
        zero_point -= 127 if symmetric else 128

    return {'op-name': 'zero_point_linear_quantize', 'zero_point': zero_point, 'scale': scale,
            'signed': signed, 'symmetric': symmetric, 'bits': 8}


def quantize_params(params, zero_point, scale, signed, symmetric):
    min = (((-127 if symmetric else -128) if signed else 0) - zero_point) * scale
    max = ((127 if signed else 255) - zero_point) * scale

    params = np.clip(params, min, max)
    return np.floor((params - min) / scale).astype(np.int8 if signed else np.uint8)


def quantize_bias(params, scale):
    return np.floor(params / scale).astype(np.int32)


def is_conv_param(tensor):
    return all(op.type in _CONV_OPS for op in tensor.consumers)


def is_conv_bias(tensor):
    assert len(tensor.consumers) == 1
    return len(tensor.consumers) == 1 and tensor.consumer.type in _CONV_OPS and len(tensor.consumer.inputs) > 2 and \
           tensor is tensor.consumer.inputs[2]


def main(args):
    reader = Reader(infer_shapes=False)
    model = reader(args.model)

    stats_path = args.statistics or os.path.join(args.model, 'graph.stats')
    if not os.path.exists(stats_path):
        print("Could not find statistics file '{}'".format(stats_path))
        return -1

    with open(stats_path, 'r') as file:
        stats = json.load(file)

    for tensor in model.tensors:
        stat = stats.get(tensor.name)
        if stat is not None:

            if args.percentile is not None:
                lo = max(stat['mean'] - args.percentile * stat['std'], stat['min'])
                hi = min(stat['mean'] + args.percentile * stat['std'], stat['max'])
            else:
                lo = stat['min']
                hi = stat['max']

            tensor.quant = make_quantization(lo, hi, args.signed, args.symmetric)

            if tensor.data is not None:
                tensor.data = quantize_params(tensor.data, tensor.quant['zero_point'], tensor.quant['scale'],
                                              args.signed, args.symmetric)

    if args.wide_bias:
        for tensor in model.tensors:
            if len(tensor.quant) > 0 and tensor.data is not None and is_conv_bias(tensor):
                conv = tensor.consumer
                tensor.quant['bits'] = 32
                tensor.quant['zero_point'] = 0
                tensor.quant['scale'] = conv.inputs[0].quant['scale'] * conv.inputs[1].quant['scale']
                tensor.data = quantize_bias(tensor.data, tensor.quant['scale'])

    writer = Writer()
    writer(model, args.output)
    return 0


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('model', type=str,
                        help='The model to quantize')
    parser.add_argument('--statistics', type=str, default=None,
                        help='The tensor statistics to use for quantization')
    parser.add_argument('--output', type=str, required=True,
                        help='The path of the output model')
    parser.add_argument('--signed', action='store_true',
                        help='Whether to generate signed int8 quantized values instead of uint8')
    parser.add_argument('--symmetric', action='store_true',
                        help='Whether to quantize symmetrically and force zero-point to 0')
    parser.add_argument('--wide-bias', action='store_true',
                        help='Whether to quantize biases into int32 values')
    parser.add_argument('--percentile', type=float, default=None,
                        help='Define ranges with approximate normal distribution percentiles;'
                             'provide number of standard deviations from mean to be used')
    exit(main(parser.parse_args()))



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys


def set_stdin_to_binary():
    if sys.version_info >= (3, 0):
        sys.stdin = sys.stdin.buffer
    elif sys.platform == 'win32':
        import os, msvcrt
        msvcrt.setmode(sys.stdin.fileno(), os.O_BINARY)


def set_stdout_to_binary():
    if sys.version_info >= (3, 0):
        sys.stdout = sys.stdout.buffer
    elif sys.platform == 'win32':
        import os, msvcrt
        msvcrt.setmode(sys.stdout.fileno(), os.O_BINARY)


def is_stdin_piped():
    return not sys.stdin.isatty()


def is_stdout_piped():
    return not sys.stdout.isatty()
